#pragma once

#include <map>
#include <string>

#include "ProviderSegmentBase.h"
#include "detail/AronTyped.h"
#include "detail/EntityContainerBase.h"
#include "detail/MaxHistorySize.h"


namespace armarx::armem::base
{

    /**
     * @brief Data of a core segment containing multiple provider segments.
     */
    template <class _ProviderSegmentT, class _Derived>
    class CoreSegmentBase :
        public detail::EntityContainerBase<_ProviderSegmentT, typename _ProviderSegmentT::EntityT, _Derived>,
        public detail::MaxHistorySize,
        public detail::AronTyped
    {
        using Base = detail::EntityContainerBase<_ProviderSegmentT, typename _ProviderSegmentT::EntityT, _Derived>;

    public:

        using typename Base::DerivedT;
        using typename Base::ContainerT;

        using ProviderSegmentT = _ProviderSegmentT;
        using EntityT = typename ProviderSegmentT::EntityT;
        using EntitySnapshotT = typename EntityT::EntitySnapshotT;
        using EntityInstanceT = typename EntitySnapshotT::EntityInstanceT;

        struct UpdateResult
        {
            armarx::armem::UpdateType coreSegmentUpdateType;
            armarx::armem::UpdateType providerSegmentUpdateType;
            armarx::armem::UpdateType entityUpdateType;
            MemoryID id;
            std::vector<EntitySnapshotT> removedSnapshots;

            UpdateResult() = default;
            UpdateResult(const typename ProviderSegmentT::UpdateResult& c) :
                providerSegmentUpdateType(c.providerSegmentUpdateType),
                entityUpdateType(c.entityUpdateType),
                id(c.id),
                removedSnapshots(c.removedSnapshots)
            {}
        };

    public:

        CoreSegmentBase()
        {
        }
        CoreSegmentBase(const std::string& name, aron::typenavigator::ObjectNavigatorPtr aronType = nullptr) :
            CoreSegmentBase(MemoryID().withCoreSegmentName(name), aronType)
        {
        }
        CoreSegmentBase(const std::string& name, const MemoryID& parentID, aron::typenavigator::ObjectNavigatorPtr aronType = nullptr) :
            CoreSegmentBase(parentID.withProviderSegmentName(name), aronType)
        {
        }
        CoreSegmentBase(const MemoryID& id, aron::typenavigator::ObjectNavigatorPtr aronType = nullptr) :
            Base(id),
            AronTyped(aronType)
        {
        }

        CoreSegmentBase(const CoreSegmentBase& other) = default;
        CoreSegmentBase(CoreSegmentBase&& other) = default;
        CoreSegmentBase& operator=(const CoreSegmentBase& other) = default;
        CoreSegmentBase& operator=(CoreSegmentBase&& other) = default;


        inline const std::string& name() const
        {
            return this->id().coreSegmentName;
        }
        inline std::string& name()
        {
            return const_cast<std::string&>(const_cast<const CoreSegmentBase*>(this)->name());
        }


        inline const std::map<std::string, ProviderSegmentT>& providerSegments() const
        {
            return this->_container;
        }
        inline std::map<std::string, ProviderSegmentT>& providerSegments()
        {
            return const_cast<std::map<std::string, ProviderSegmentT>&>(const_cast<const CoreSegmentBase*>(this)->providerSegments());
        }


        bool hasProviderSegment(const std::string& name) const
        {
            return this->_container.count(name) > 0;
        }

        ProviderSegmentT& getProviderSegment(const std::string& name)
        {
            return const_cast<ProviderSegmentT&>(const_cast<const CoreSegmentBase*>(this)->getProviderSegment(name));
        }

        const ProviderSegmentT& getProviderSegment(const std::string& name) const
        {
            auto it = this->_container.find(name);
            if (it != this->_container.end())
            {
                return it->second;
            }
            else
            {
                throw armem::error::MissingEntry::create<ProviderSegmentT>(name, *this);
            }
        }

        using Base::getEntity;
        const EntityT& getEntity(const MemoryID& id) const override
        {
            this->_checkContainerName(id.coreSegmentName, this->getKeyString());
            return getProviderSegment(id.providerSegmentName).getEntity(id);
        }

        const EntityT* findEntity(const MemoryID& id) const override
        {
            this->_checkContainerName(id.coreSegmentName, this->getKeyString());
            if (id.hasProviderSegmentName())
            {
                return getProviderSegment(id.providerSegmentName).findEntity(id);
            }
            else
            {
                for (const auto& [_, providerSegment] : this->_container)
                {
                    if (auto entity = providerSegment.findEntity(id))
                    {
                        return entity;
                    }
                }
                return nullptr;
            }
        }

        /**
         * @brief Updates an entity's history.
         *
         * Missing entity entries are added before updating.
         */
        UpdateResult update(const EntityUpdate& update)
        {
            this->_checkContainerName(update.entityID.coreSegmentName, this->name());

            ProviderSegmentT* provSeg;
            UpdateType updateType = UpdateType::InsertedNew;

            auto it = this->_container.find(update.entityID.providerSegmentName);
            if (it == this->_container.end())
            {
                if (_addMissingProviderSegmentDuringUpdate)
                {
                    // Insert into map.
                    provSeg = &addProviderSegment(update.entityID.providerSegmentName);
                    provSeg->setMaxHistorySize(_maxHistorySize);
                    updateType = UpdateType::InsertedNew;
                }
                else
                {
                    throw error::MissingEntry::create<EntitySnapshotT>(update.entityID.providerSegmentName, *this);
                }
            }
            else
            {
                provSeg = &it->second;
                updateType = UpdateType::UpdatedExisting;
            }

            // Update entry.
            UpdateResult ret(provSeg->update(update));
            ret.coreSegmentUpdateType = updateType;
            return ret;
        }

        void append(const _Derived& m)
        {
            ARMARX_INFO << "CoreSegment: Merge name '" << m.name() << "' into '" << name() << "'";

            for (const auto& [k, s] : m.container())
            {
                if (const auto& it = this->_container.find(k); it != this->_container.end())
                {
                    // segment already exists
                    it->second.append(s);
                }
                else
                {
                    auto wms = this->_container.emplace(std::make_pair(k, this->id().withProviderSegmentName(k)));
                    wms.first->second.append(s);
                }
            }
        }

        /**
         * @brief Add an empty provider segment with the given name and optional provider segment type.
         * @param name The segment name.
         * @param providerSegmentType The provider type. If nullptr, the core segment type is used.
         * @return The added provider segment.
         */
        ProviderSegmentT& addProviderSegment(const std::string& name, aron::typenavigator::ObjectNavigatorPtr providerSegmentType = nullptr)
        {
            aron::typenavigator::ObjectNavigatorPtr type = providerSegmentType ? providerSegmentType : this->aronType();
            return addProviderSegment(ProviderSegmentT(name, type));
        }

        /// Copy and insert a provider segment.
        ProviderSegmentT& addProviderSegment(const ProviderSegmentT& providerSegment)
        {
            return addProviderSegment(ProviderSegment(providerSegment));
        }

        /// Move and insert a provider segment.
        ProviderSegmentT& addProviderSegment(ProviderSegmentT&& providerSegment)
        {
            if (hasProviderSegment(providerSegment.name()))
            {
                throw armem::error::ContainerEntryAlreadyExists(
                    providerSegment.getLevelName(), providerSegment.name(), getLevelName(), this->getKeyString());
            }

            auto it = this->_container.emplace(providerSegment.name(), std::move(providerSegment)).first;
            it->second.id().setCoreSegmentID(this->id());
            it->second.setMaxHistorySize(_maxHistorySize);
            return it->second;
        }


        /**
         * @brief Sets the maximum history size of entities in this segment.
         * This affects all current entities as well as new ones.
         * @see Entity::setMaxHistorySize()
         */
        void setMaxHistorySize(long maxSize) override
        {
            MaxHistorySize::setMaxHistorySize(maxSize);
            for (auto& [name, seg] : this->_container)
            {
                seg.setMaxHistorySize(maxSize);
            }
        }

        virtual bool equalsDeep(const CoreSegmentBase& other) const
        {
            //std::cout << "CoreSegment::equalsDeep" << std::endl;
            if (this->size() != other.size())
            {
                return false;
            }
            for (const auto& [key, provider] : this->_container)
            {
                if (not other.hasProviderSegment(key))
                {
                    return false;
                }
                if (not provider.equalsDeep(other.getProviderSegment(key)))
                {
                    return false;
                }
            }
            return true;
        }

        std::string getLevelName() const override
        {
            return "core segment";
        }

        std::string getKeyString() const override
        {
            return this->name();
        }


    protected:

        virtual void _copySelf(DerivedT& other) const override
        {
            Base::_copySelf(other);
            other.aronType() = _aronType;
            other._addMissingProviderSegmentDuringUpdate = _addMissingProviderSegmentDuringUpdate;
        }
        virtual void _copySelfEmpty(DerivedT& other) const override
        {
            Base::_copySelfEmpty(other);
            other.aronType() = _aronType;
            other._addMissingProviderSegmentDuringUpdate = _addMissingProviderSegmentDuringUpdate;
        }


    private:

        bool _addMissingProviderSegmentDuringUpdate = true;

    };

}
