#pragma once

#include <map>
#include <string>

#include <SimoxUtility/algorithm/get_map_keys_values.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "../../core/Time.h"
#include "../../core/MemoryID.h"

#include "EntitySnapshotBase.h"
#include "detail/MaxHistorySize.h"
#include "detail/MemoryContainerBase.h"


namespace armarx::armem::base
{

    /**
     * @brief An entity over a period of time.
     *
     * An entity should be a physical thing or abstract concept existing
     * (and potentially evolving) over some time.
     *
     * Examples are:
     * - objects (the green box)
     * - agents (robot, human)
     * - locations (frige, sink)
     * - grasp affordances (general, or for a specific object)
     * - images
     * - point clouds
     * - other sensory values
     *
     * At each point in time (`EntitySnapshot`), the entity can have a
     * (potentially variable) number of instances (`EntityInstance`),
     * each containing a single `AronData` object of a specific `AronType`.
     */
    template <class _EntitySnapshotT, class _Derived>
    class EntityBase :
        public detail::MemoryContainerBase<std::map<Time, _EntitySnapshotT>, _Derived>,
        public detail::MaxHistorySize
    {
        using Base = detail::MemoryContainerBase<std::map<Time, _EntitySnapshotT>, _Derived>;

    public:

        using typename Base::DerivedT;
        using typename Base::ContainerT;

        using EntitySnapshotT = _EntitySnapshotT;
        using EntityInstanceT = typename EntitySnapshotT::EntityInstanceT;

        struct UpdateResult
        {
            armarx::armem::UpdateType entityUpdateType;
            MemoryID id;
            std::vector<EntitySnapshotT> removedSnapshots;

            UpdateResult() = default;
        };

    public:

        EntityBase()
        {
        }
        EntityBase(const std::string& name, const MemoryID& parentID = {}) :
            EntityBase(parentID.withEntityName(name))
        {
        }
        EntityBase(const MemoryID& id) :
            Base(id)
        {
        }


        EntityBase(const EntityBase& other) = default;
        EntityBase(EntityBase&& other) = default;
        EntityBase& operator=(const EntityBase& other) = default;
        EntityBase& operator=(EntityBase&& other) = default;


        virtual bool equalsDeep(const EntityBase& other) const
        {
            //std::cout << "Entity::equalsDeep" << std::endl;
            if (this->size() != other.size())
            {
                return false;
            }
            for (const auto& [key, snapshot] : this->_container)
            {
                if (not other.hasSnapshot(key))
                {
                    return false;
                }
                if (not snapshot.equalsDeep(other.getSnapshot(key)))
                {
                    return false;
                }
            }
            return true;
        }


        inline const std::string& name() const
        {
            return this->id().entityName;
        }
        inline std::string& name()
        {
            return const_cast<std::string&>(const_cast<const EntityBase*>(this)->name());
        }


        inline const std::map<Time, EntitySnapshotT>& history() const
        {
            return this->_container;
        }
        inline std::map<Time, EntitySnapshotT>& history()
        {
            return const_cast<std::map<Time, EntitySnapshotT>&>(const_cast<const EntityBase*>(this)->history());
        }


        /**
         * @brief Indicates whether a history entry for the given time exists.
         */
        virtual bool hasSnapshot(const Time& time)
        {
            return const_cast<const EntityBase*>(this)->hasSnapshot(time);
        }

        virtual bool hasSnapshot(const Time& time) const
        {
            return this->_container.count(time) > 0;
        }

        /**
         * @brief Get the latest timestamp.
         * @throw `armem::error::EntityHistoryEmpty` If the history is empty.
         */
        Time getLatestTimestamp()
        {
            return const_cast<const EntityBase*>(this)->getLatestTimestamp();
        }

        Time getLatestTimestamp() const
        {
            return getLatestItem().first;
        }

        /**
         * @brief Get all timestamps in the history.
         */
        virtual std::vector<Time> getTimestamps()
        {
            return const_cast<const EntityBase*>(this)->getTimestamps();
        }

        virtual std::vector<Time> getTimestamps() const
        {
            return simox::alg::get_keys(this->_container);
        }


        /**
         * @brief Get a snapshot.
         * @param time The time.
         * @return The snapshot, if it exists.
         *
         * @throws `armem::error::MissingEntry` If there is no such entry.
         */
        virtual EntitySnapshotT& getSnapshot(const Time& time)
        {
            return const_cast<EntitySnapshotT&>(const_cast<const EntityBase*>(this)->getSnapshot(time));
        }

        virtual const EntitySnapshotT& getSnapshot(const Time& time) const
        {
            auto it = this->_container.find(time);
            if (it != this->_container.end())
            {
                return it->second;
            }
            else
            {
                throw armem::error::MissingEntry::create<EntitySnapshotT>(toDateTimeMilliSeconds(time), *this);
            }
        }

        EntitySnapshotT& getSnapshot(const MemoryID& id)
        {
            return const_cast<EntitySnapshotT&>(const_cast<const EntityBase*>(this)->getSnapshot(id));
        }

        const EntitySnapshotT& getSnapshot(const MemoryID& id) const
        {
            this->_checkContainerName(id.entityName, this->name());
            return getSnapshot(id.timestamp);
        }

        /**
         * @brief Return the snapshot with the most recent timestamp.
         * @return The latest snapshot.
         * @throw `armem::error::EntityHistoryEmpty` If the history is empty.
         */
        EntitySnapshotT& getLatestSnapshot()
        {
            return const_cast<EntitySnapshotT&>(const_cast<const EntityBase*>(this)->getLatestSnapshot());
        }

        const EntitySnapshotT& getLatestSnapshot() const
        {
            return getLatestItem().second;
        }

        /**
         * @brief Return the lastest snapshot before or at time.
         * @param time The time.
         * @return The latest snapshot.
         * @throw `armem::error::EntityHistoryEmpty` If the history is empty.
         * @throw `armem::error::MissingEntry` If no such snapshot
         */
        virtual const EntitySnapshotT& getLatestSnapshotBeforeOrAt(const Time& time) const
        {
            // first element equal or greater
            typename std::map<Time, EntitySnapshotT>::const_iterator refIt = this->_container.upper_bound(time);

            if (refIt == this->_container.begin())
            {
                if (refIt->first == time)
                {
                    return refIt->second;
                }
                throw error::MissingEntry::create<EntitySnapshotT>(toDateTimeMilliSeconds(time), *this);
            }

            // last element less than
            typename std::map<Time, EntitySnapshotT>::const_iterator refItPrev = std::prev(refIt);

            // ... or we return the element before if possible
            if (refItPrev != this->_container.begin())
            {
                return refItPrev->second;
            }

            throw error::MissingEntry::create<EntitySnapshotT>(toDateTimeMilliSeconds(time), *this);
        }

        /**
         * @brief Return the lastest snapshot before time.
         * @param time The time.
         * @return The latest snapshot.
         * @throw `armem::error::EntityHistoryEmpty` If the history is empty.
         * @throw `armem::error::MissingEntry` If no such snapshot
         */
        virtual const EntitySnapshotT& getLatestSnapshotBefore(const Time& time) const
        {
            // first element equal or greater
            typename std::map<Time, EntitySnapshotT>::const_iterator refIt = this->_container.upper_bound(time);

            if (refIt == this->_container.begin())
            {
                throw error::MissingEntry::create<EntitySnapshotT>(toDateTimeMilliSeconds(time), *this);
            }

            // last element less than
            typename std::map<Time, EntitySnapshotT>::const_iterator refItPrev = std::prev(refIt);

            // we return the element before if possible
            if (refItPrev != this->_container.begin())
            {
                return refItPrev->second;
            }

            throw error::MissingEntry::create<EntitySnapshotT>(toDateTimeMilliSeconds(time), *this);
        }

        /**
         * @brief Return all snapshots before or at time.
         * @param time The time.
         * @return The latest snapshots.
         * @throw `armem::error::EntityHistoryEmpty` If the history is empty.
         * @throw `armem::error::MissingEntry` If no such snapshots
         */
        virtual const std::vector<std::reference_wrapper<const EntitySnapshotT>> getSnapshotsBefore(const Time& time) const
        {
            std::vector<std::reference_wrapper<const EntitySnapshotT>> ret;
            const EntitySnapshotT& latest = getLatestSnapshotBefore(time);

            for (const auto& [timestamp, snapshot] : this->_container)
            {
                ret.emplace_back(snapshot);
                if (timestamp == latest.id().timestamp)
                {
                    break;
                }
            }
            return ret;
        }

        /**
         * @brief Return first snapshot after or at time.
         * @param time The time.
         * @return The first snapshot.
         * @throw `armem::error::EntityHistoryEmpty` If the history is empty.
         * @throw `armem::error::MissingEntry` If no such snapshot
         */
        virtual const EntitySnapshotT& getFirstSnapshotAfterOrAt(const Time& time) const
        {
            // first element equal or greater
            typename std::map<Time, EntitySnapshotT>::const_iterator refIt = this->_container.upper_bound(time);

            if (refIt == this->_container.end())
            {
                throw error::MissingEntry::create<EntitySnapshotT>(toDateTimeMilliSeconds(time), *this);
            }
            return refIt->second;
        }

        /**
         * @brief Return first snapshot after time.
         * @param time The time.
         * @return The first snapshot.
         * @throw `armem::error::EntityHistoryEmpty` If the history is empty.
         * @throw `armem::error::MissingEntry` If no such snapshot
         */
        virtual const EntitySnapshotT& getFirstSnapshotAfter(const Time& time) const
        {
            // first element equal or greater
            typename std::map<Time, EntitySnapshotT>::const_iterator refIt = this->_container.upper_bound(time);

            if (refIt == this->_container.end())
            {
                throw error::MissingEntry::create<EntitySnapshotT>(toDateTimeMilliSeconds(time), *this);
            }

            if (refIt->first > time)
            {
                return refIt->second;
            }

            // first element greater than
            typename std::map<Time, EntitySnapshotT>::const_iterator refItNext = std::next(refIt);

            if (refItNext != this->_container.end())
            {
                return refItNext->second;
            }

            throw error::MissingEntry::create<EntitySnapshotT>(toDateTimeMilliSeconds(time), *this);
        }


        /**
         * @brief Add the given update to this entity's history.
         * @param update The update.
         * @return The snapshot ID of the update.
         */
        UpdateResult update(const EntityUpdate& update)
        {
            this->_checkContainerName(update.entityID.entityName, this->name());
            UpdateResult ret;

            EntitySnapshotT* snapshot;

            auto it = this->_container.find(update.timeCreated);
            if (it == this->_container.end())
            {
                // Insert into history.
                snapshot = &addSnapshot(update.timeCreated);
                ret.removedSnapshots = truncate();
                ret.entityUpdateType = UpdateType::InsertedNew;
            }
            else
            {
                snapshot = &it->second;
                ret.entityUpdateType = UpdateType::UpdatedExisting;
            }
            // Update entry.
            snapshot->update(update, this->id());
            ret.id = snapshot->id();

            return ret;
        }

        void append(const DerivedT& m)
        {
            ARMARX_INFO << "Entity: Merge name '" << m.name() << "' into '" << name() << "'";

            for (const auto& [k, s] : m.container())
            {
                if (const auto& it = this->_container.find(k); it != this->_container.end())
                {
                    // segment already exists
                    // We assume that a snapshot does not change, so ignore
                }
                else
                {
                    auto snapshot = s.copy();
                    snapshot.id() = this->id().withTimestamp(k); // update id (e.g. memory name) if necessary
                    this->_container.insert(std::make_pair(k, std::move(snapshot)));
                }
            }
        }

        /**
         * @brief Add a single snapshot with data.
         * @param snapshot The snapshot.
         * @return The stored snapshot.
         */
        EntitySnapshotT& addSnapshot(const EntitySnapshotT& snapshot)
        {
            return addSnapshot(EntitySnapshotT(snapshot));
        }

        EntitySnapshotT& addSnapshot(EntitySnapshotT&& snapshot)
        {
            auto it = this->_container.emplace(snapshot.time(), std::move(snapshot)).first;
            it->second.id().setEntityID(this->id());
            return it->second;
        }

        EntitySnapshotT& addSnapshot(const Time& timestamp)
        {
            return addSnapshot(EntitySnapshotT(timestamp));
        }


        /**
         * @brief Sets the maximum history size.
         *
         * The current history is truncated if necessary.
         */
        void setMaxHistorySize(long maxSize) override
        {
            MaxHistorySize::setMaxHistorySize(maxSize);
            truncate();
        }

        std::string getKeyString() const override
        {
            return this->id().entityName;
        }
        std::string getLevelName() const override
        {
            return "entity";
        }


    protected:

        /// If maximum size is set, ensure `history`'s is not higher.
        std::vector<EntitySnapshotT> truncate()
        {
            std::vector<EntitySnapshotT> removedElements;
            if (_maxHistorySize >= 0)
            {
                while (this->_container.size() > size_t(_maxHistorySize))
                {
                    removedElements.push_back(std::move(this->_container.begin()->second));
                    this->_container.erase(this->_container.begin());
                }
                ARMARX_CHECK_LESS_EQUAL(this->_container.size(), _maxHistorySize);
            }
            return removedElements;
        }

        /**
         * @brief Return the snapshot with the most recent timestamp.
         * @return The latest snapshot.
         * @throw `armem::error::EntityHistoryEmpty` If the history is empty.
         */
        virtual const typename std::map<Time, EntitySnapshotT>::value_type& getLatestItem() const
        {
            if (this->_container.empty())
            {
                throw armem::error::EntityHistoryEmpty(name(), "when getting the latest snapshot.");
            }
            return *this->_container.rbegin();
        }

    };

}
