#include "Segment.h"

#include <SimoxUtility/algorithm/get_map_keys_values.h>
#include <sstream>

#include <ArmarXCore/core/time/TimeUtil.h>
#include "ArmarXCore/core/logging/Logging.h"

#include "RobotAPI/libraries/aron/common/aron_conversions.h"

#include <RobotAPI/libraries/armem/core/aron_conversions.h>
#include <RobotAPI/libraries/armem/core/workingmemory/visitor.h>
#include "RobotAPI/libraries/armem/core/MemoryID.h"
#include <RobotAPI/libraries/armem/client/Writer.h>
#include <RobotAPI/libraries/armem/client/query/Builder.h>
#include <RobotAPI/libraries/armem/client/query/query_fns.h>
#include <RobotAPI/libraries/armem/server/MemoryToIceAdapter.h>

#include "RobotAPI/libraries/armem_robot/robot_conversions.h"
#include <RobotAPI/libraries/armem_robot/aron/Robot.aron.generated.h>
#include <RobotAPI/libraries/armem_robot/aron_conversions.h>


namespace armarx::armem::server::obj::articulated_object_class
{

    Segment::Segment(armem::server::MemoryToIceAdapter& memoryToIceAdapter, std::mutex& memoryMutex) :
        iceMemory(memoryToIceAdapter),
        memoryMutex(memoryMutex)
    {
        Logging::setTag("ArticulatedObjectInstanceSegment");
    }

    Segment::~Segment() = default;

    void Segment::defineProperties(armarx::PropertyDefinitionsPtr defs, const std::string& prefix)
    {
        defs->optional(p.coreClassSegmentName, prefix + "CoreSegmentName", "Name of the object instance core segment.");
        defs->optional(p.maxHistorySize, prefix + "MaxHistorySize", "Maximal size of object poses history (-1 for infinite).");

        defs->optional(p.objectsPackage, prefix + "ObjectsPackage", "Name of the objects package to load from.");
        defs->optional(p.loadFromObjectsPackage, prefix + "LoadFromObjectsPackage",
                       "If true, load the objects from the objects package on startup.");

    }

    void Segment::init()
    {
        ARMARX_CHECK_NOT_NULL(iceMemory.workingMemory);

        coreSegment = &iceMemory.workingMemory->addCoreSegment(p.coreClassSegmentName, arondto::RobotDescription::toAronType());
        coreSegment->setMaxHistorySize(p.maxHistorySize);

        if (p.loadFromObjectsPackage)
        {
            loadByObjectFinder(p.objectsPackage);
        }
    }

    void Segment::connect(viz::Client arviz)
    {
        // this->visu = std::make_unique<Visu>(arviz, *this);
    }

    void Segment::loadByObjectFinder(const std::string& package)
    {
        ObjectFinder finder(package);

        const auto knownArticulatedObjectDescriptions = finder.findAllArticulatedObjectsByDataset(true);
        ARMARX_DEBUG << "Found " << knownArticulatedObjectDescriptions.size() << " articulated objects";

        loadObjectsIntoMemory(knownArticulatedObjectDescriptions, package);
    }

    void Segment::loadObjectsIntoMemory(const std::unordered_map<std::string, std::vector<armem::articulated_object::ArticulatedObjectDescription>>& datasets, const std::string& package)
    {
        const Time now = TimeUtil::GetTime();

        const MemoryID providerID = coreSegment->id().withProviderSegmentName(package);
        coreSegment->addProviderSegment(providerID.providerSegmentName);

        // ARMARX_INFO << "Loading up to " << infos.size() << " object classes from '"
        //             << objectFinder.getPackageName() << "' ...";
        Commit commit;

        for (const auto& [datasetName, descriptions] : datasets)
        {

            for (const armem::articulated_object::ArticulatedObjectDescription& desc : descriptions)
            {
                EntityUpdate& update = commit.updates.emplace_back();
                update.entityID = providerID.withEntityName(desc.name);
                update.timeArrived = update.timeCreated = update.timeSent = now;

                arondto::RobotDescription aronRobotDescription;
                toAron(aronRobotDescription, desc);
                // TODO toAron(aronRobotDescription.timestamp, now);

                update.instancesData = { aronRobotDescription.toAron()};
            }

            ARMARX_INFO << "Loaded " << commit.updates.size() << " articulated object classes from '"
                        << package << "' in dataset '" << datasetName << "'.";
        }
        iceMemory.commit(commit);
    }



    std::unordered_map<armem::MemoryID, ::armarx::armem::articulated_object::ArticulatedObjectDescription> Segment::getKnownObjectClasses() const
    {
        std::unordered_map<armem::MemoryID, ::armarx::armem::articulated_object::ArticulatedObjectDescription> objects;

        for (const auto& [_, provSeg] : iceMemory.workingMemory->getCoreSegment(p.coreClassSegmentName))
        {
            for (const auto& [name, entity] :  provSeg.entities())
            {
                for (const auto& snapshot : simox::alg::get_values(entity.history()))
                {
                    const auto& entityInstance = snapshot.getInstance(0);
                    const auto description = robot::convertRobotDescription(entityInstance);

                    if (not description)
                    {
                        ARMARX_WARNING << "Could not convert entity instance to 'RobotDescription'";
                        continue;
                    }

                    ARMARX_DEBUG << "Key is " << armem::MemoryID(snapshot.id());

                    objects.emplace(armem::MemoryID(snapshot.id()), *description);
                }
            }
        }

        ARMARX_INFO << deactivateSpam(10) <<  "Number of known articulated object classes: " << objects.size();

        return objects;
    }



    // void Segment::RemoteGui::setup(const Segment& data)
    // {
    //     using namespace armarx::RemoteGui::Client;

    //     maxHistorySize.setValue(std::max(1, int(data.p.maxHistorySize)));
    //     maxHistorySize.setRange(1, 1e6);
    //     infiniteHistory.setValue(data.p.maxHistorySize == -1);
    //     discardSnapshotsWhileAttached.setValue(data.p.discardSnapshotsWhileAttached);

    //     GridLayout grid;
    //     int row = 0;
    //     grid.add(Label("Max History Size"), {row, 0}).add(maxHistorySize, {row, 1});
    //     row++;
    //     grid.add(Label("Infinite History Size"), {row, 0}).add(infiniteHistory, {row, 1});
    //     row++;
    //     grid.add(Label("Discard Snapshots while Attached"), {row, 0}).add(discardSnapshotsWhileAttached, {row, 1});
    //     row++;

    //     group.setLabel("Data");
    //     group.addChild(grid);
    // }

    // void Segment::RemoteGui::update(Segment& data)
    // {
    //     if (infiniteHistory.hasValueChanged() || maxHistorySize.hasValueChanged()
    //         || discardSnapshotsWhileAttached.hasValueChanged())
    //     {
    //         std::scoped_lock lock(data.memoryMutex);

    //         if (infiniteHistory.hasValueChanged() || maxHistorySize.hasValueChanged())
    //         {
    //             data.p.maxHistorySize = infiniteHistory.getValue() ? -1 : maxHistorySize.getValue();
    //             if (data.coreSegment)
    //             {
    //                 data.coreSegment->setMaxHistorySize(long(data.p.maxHistorySize));
    //             }
    //         }

    //         data.p.discardSnapshotsWhileAttached = discardSnapshotsWhileAttached.getValue();
    //     }
    // }

}  // namespace armarx::armem::server::obj::articulated_object_class
