/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


// Simox
#include <SimoxUtility/algorithm/string.h>

// Header
#include "Writer.h"

// ArmarX

namespace armarx::aron::cppserializer
{
    Writer::Writer(const std::string& producerName, const std::vector<std::string>& additionalIncludesFromXMLFile) :
        CodeWriter(producerName, additionalIncludesFromXMLFile)
    {
        addToAronMethod();
        addFromAronMethod();
        addToAronTypeMethod();
    }

    void Writer::addToAronMethod()
    {
        // The toAron Serializer is visible by default
        {
            codegeneratorhelper::WriterInfoPtr toAron = codegeneratorhelper::WriterInfoPtr(new codegeneratorhelper::WriterInfo());
            toAron->methodName = "toAron";
            toAron->returnType = "armarx::aron::datanavigator::DictNavigatorPtr";
            toAron->writerClassType = "armarx::aron::dataIO::writer::NavigatorWriter";
            toAron->include = "<RobotAPI/libraries/aron/core/io/dataIO/writer/navigator/NavigatorWriter.h>";
            dataWriters.push_back(toAron);
        }
    }

    void Writer::addFromAronMethod()
    {
        // The toAron Serializer is visible by default
        {
            codegeneratorhelper::ReaderInfoPtr fromAron = codegeneratorhelper::ReaderInfoPtr(new codegeneratorhelper::ReaderInfo());
            fromAron->methodName = "fromAron";
            fromAron->argumentType = "armarx::aron::datanavigator::DictNavigatorPtr";
            fromAron->readerClassType = "armarx::aron::dataIO::reader::NavigatorReader";
            fromAron->include = "<RobotAPI/libraries/aron/core/io/dataIO/reader/navigator/NavigatorReader.h>";
            dataReaders.push_back(fromAron);
        }
        {
            codegeneratorhelper::ReaderInfoPtr fromAron = codegeneratorhelper::ReaderInfoPtr(new codegeneratorhelper::ReaderInfo());
            fromAron->methodName = "fromAron";
            fromAron->argumentType = "armarx::aron::data::AronDictPtr";
            fromAron->readerClassType = "armarx::aron::dataIO::reader::NavigatorReader";
            dataReaders.push_back(fromAron);
        }
    }

    void Writer::addToAronTypeMethod()
    {
        // The toAron Serializer is visible by default
        codegeneratorhelper::WriterInfoPtr toAronType = codegeneratorhelper::WriterInfoPtr(new codegeneratorhelper::WriterInfo());
        toAronType->methodName = "toAronType";
        toAronType->returnType = "armarx::aron::typenavigator::ObjectNavigatorPtr";
        toAronType->writerClassType = "armarx::aron::typeIO::writer::NavigatorWriter";
        toAronType->include = "<RobotAPI/libraries/aron/core/io/typeIO/writer/navigator/NavigatorWriter.h>";
        initialTypeWriters.push_back(toAronType);
    }

    void Writer::generateTypeObjects(const std::vector<codegeneratorhelper::GenerateObjectInfoPtr>& generateObjects)
    {
        for (const auto& publicGenerateObjectType : generateObjects)
        {
            const auto& nav = publicGenerateObjectType->correspondingType;

            //std::cout << "Generating: " << nav->getName() << std::endl;

            // Convert to Object type and create class object
            if (nav == nullptr)
            {
                throw error::AronException("ClassCppWriter", "generateTypeClasses", "An received public type is null. Abort due to error!");
            }

            auto serializer = std::make_shared<serializer::ObjectClassSerializer>(nav);

            CppClassPtr c = setupBasicCppClass(serializer->getFullCppTypename());
            setupMemberFields(c, serializer);

            if (nav->getExtends() != nullptr)
            {
                auto extendsSerializer = std::make_shared<serializer::ObjectClassSerializer>(nav->getExtends());
                c->addInherit(extendsSerializer->getFullCppTypename());
            }
            else
            {
                c->addInherit("public armarx::aron::cppserializer::AronCppClass");
                c->addInclude("<RobotAPI/libraries/aron/core/codegenerator/codeWriter/cpp/AronCppClass.h>");
            }

            // Add includes and guard
            c->addClassDoc("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"
                           "!!!!!!AUTOGENERATED CLASS. Please do NOT edit. Instead, edit the corresponding .xml file!!!!!!\n"
                           "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
            c->setPragmaOnceIncludeGuard(true);
            c->addInclude("<memory>");
            c->addInclude("<string>");
            c->addInclude("<vector>");
            c->addInclude("<map>");
            c->addInclude("<RobotAPI/interface/aron.h>");
            for (const std::string& s : additionalIncludes)
            {
                if (!s.empty())
                {
                    c->addInclude(s);
                }
            }

            //std::cout << "Generate Type classes" << std::endl;
            generateInnerTypeObjects(c, publicGenerateObjectType->nestedObjects);

            // ctor
            c->addCtor(serializer->toCtor(c->getName()));

            // Generic methods
            //std::cout << "Generate equals method" << std::endl;
            CppMethodPtr equals = serializer->toEqualsMethod();
            c->addMethod(equals);

            //std::cout << "Generate reset method" << std::endl;
            CppMethodPtr resetHard = serializer->toResetHardMethod();
            c->addMethod(resetHard);

            //std::cout << "Generate init method" << std::endl;
            CppMethodPtr resetSoft = serializer->toResetSoftMethod();
            c->addMethod(resetSoft);

            //std::cout << "Generate writeInit method" << std::endl;
            CppMethodPtr writeType = serializer->toWriteTypeMethod();
            c->addMethod(writeType);

            //std::cout << "Generate write method" << std::endl;
            CppMethodPtr write = serializer->toWriteMethod();
            c->addMethod(write);

            //std::cout << "Generate read method" << std::endl;
            CppMethodPtr read = serializer->toReadMethod();
            c->addMethod(read);

            // Writermethods
            for (const codegeneratorhelper::WriterInfoPtr& info : dataWriters)
            {
                if (!info->include.empty())
                {
                    c->addInclude(info->include);
                }
                CppMethodPtr convert = serializer->toSpecializedDataWriterMethod(info->returnType, info->methodName, info->writerClassType, "armarx::aron::datanavigator::DictNavigator::DynamicCast");
                c->addMethod(convert);
            }

            //std::cout << "set member vars" << std::endl;

            // Add methods to set the member variables
            for (const codegeneratorhelper::ReaderInfoPtr& info : dataReaders)
            {
                if (!info->include.empty())
                {
                    c->addInclude(info->include);
                }
                CppMethodPtr convert = serializer->toSpecializedDataReaderMethod(info->argumentType, info->methodName, info->readerClassType);
                c->addMethod(convert);
            }

            //std::cout << "type writer methods" << std::endl;

            // Typewritermethods
            for (const codegeneratorhelper::WriterInfoPtr& info : initialTypeWriters)
            {
                if (!info->include.empty())
                {
                    c->addInclude(info->include);
                }
                CppMethodPtr convert = serializer->toSpecializedTypeWriterMethod(info->returnType, info->methodName, info->writerClassType, "armarx::aron::typenavigator::ObjectNavigator::DynamicCast");
                c->addMethod(convert);
            }

            //std::cout << "push back" << std::endl;

            typeClasses.push_back(c);
        }
    }

    void Writer::generateTypeIntEnums(const std::vector<codegeneratorhelper::GenerateIntEnumInfoPtr>& generateIntEnums)
    {
        for (const auto& publicGenerateIntEnumType : generateIntEnums)
        {
            const auto& nav = publicGenerateIntEnumType->correspondingType;
            if (nav == nullptr)
            {
                throw error::AronException("ClassCppWriter", "generateTypeIntEnums", "An received public type is null. Abort due to error!");
            }

            auto serializer = std::make_shared<serializer::IntEnumClassSerializer>(nav);

            CppClassPtr c = setupBasicCppClass(serializer->getFullCppTypename());
            setupMemberFields(c, serializer);

            // Add includes and guard
            c->addClassDoc("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"
                           "!!!!!!AUTOGENERATED CLASS. Please do NOT edit. Instead, edit the corresponding .xml file!!!!!!\n"
                           "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
            c->setPragmaOnceIncludeGuard(true);
            c->addInclude("<memory>");
            c->addInclude("<string>");
            c->addInclude("<vector>");
            c->addInclude("<map>");
            c->addInclude("<RobotAPI/interface/aron.h>");

            c->addInherit("virtual public armarx::aron::cppserializer::AronCppClass");
            c->addInclude("<RobotAPI/libraries/aron/core/codegenerator/codeWriter/cpp/AronCppClass.h>");

            // ctor
            c->addCtor(serializer->toCtor(c->getName()));
            c->addCtor(serializer->toCopyCtor(c->getName()));
            c->addCtor(serializer->toInnerEnumCtor(c->getName()));

            // Specific methods
            CppEnumPtr enumrepresentation = serializer->toInnerEnumDefinition();
            c->addInnerEnum(enumrepresentation);

            CppMethodPtr toString = serializer->toToStringMethod();
            c->addMethod(toString);

            CppMethodPtr fromString = serializer->toFromStringMethod();
            c->addMethod(fromString);

            CppMethodPtr intConversion = serializer->toIntMethod();
            c->addMethod(intConversion);

            CppMethodPtr enumAssignment = serializer->toEnumAssignmentMethod();
            c->addMethod(enumAssignment);

            CppMethodPtr enumAssignment2 = serializer->toCopyAssignmentMethod();
            c->addMethod(enumAssignment2);

            CppMethodPtr enumAssignment3 = serializer->toIntAssignmentMethod();
            c->addMethod(enumAssignment3);

            // Generic methods
            //std::cout << "Generate equals method" << std::endl;
            CppMethodPtr equals = serializer->toEqualsMethod();
            c->addMethod(equals);

            //std::cout << "Generate reset method" << std::endl;
            CppMethodPtr resetHard = serializer->toResetHardMethod();
            c->addMethod(resetHard);

            //std::cout << "Generate init method" << std::endl;
            CppMethodPtr resetSoft = serializer->toResetSoftMethod();
            c->addMethod(resetSoft);

            //std::cout << "Generate writeInit method" << std::endl;
            CppMethodPtr writeInitialType = serializer->toWriteTypeMethod();
            c->addMethod(writeInitialType);

            //std::cout << "Generate write method" << std::endl;
            CppMethodPtr write = serializer->toWriteMethod();
            c->addMethod(write);

            //std::cout << "Generate read method" << std::endl;
            CppMethodPtr read = serializer->toReadMethod();
            c->addMethod(read);

            typeClasses.push_back(c);
        }
    }

    void Writer::generateInnerTypeObjects(CppClassPtr& classToAdd, const std::map<std::string, codegeneratorhelper::GenerateObjectInfoPtr>& localGenerateTypes)
    {
        // setup inner classes
        for (const auto& [_, localGenerateObjectType] : localGenerateTypes)
        {
            const auto& nav = localGenerateObjectType->correspondingType;

            auto serializer = std::make_shared<serializer::ObjectClassSerializer>(nav);

            CppClassPtr c = setupBasicCppClass(serializer->getFullCppTypename());
            setupMemberFields(c, serializer);

            if (nav->getExtends() != nullptr)
            {
                auto extendsSerializer = std::make_shared<serializer::ObjectClassSerializer>(nav->getExtends());
                c->addInherit(extendsSerializer->getFullCppTypename());
            }
            else
            {
                c->addInherit("virtual public armarx::aron::cppserializer::AronCppClass");
            }

            CppMethodPtr equals = serializer->toEqualsMethod();
            c->addMethod(equals);

            CppMethodPtr resetHard = serializer->toResetHardMethod();
            c->addMethod(resetHard);

            CppMethodPtr resetSoft = serializer->toResetSoftMethod();
            c->addMethod(resetSoft);

            CppMethodPtr writeInitialType = serializer->toWriteTypeMethod();
            c->addMethod(writeInitialType);

            CppMethodPtr write = serializer->toWriteMethod();
            c->addMethod(write);

            CppMethodPtr read = serializer->toReadMethod();
            c->addMethod(read);

            generateInnerTypeObjects(c, localGenerateObjectType->nestedObjects);
            classToAdd->addInnerClass(c);
        }
    }

    CppClassPtr Writer::setupBasicCppClass(const std::string& cppTypename) const
    {
        std::vector<std::string> split = simox::alg::split(cppTypename, "::");
        if (split.size() < 1)
        {
            throw error::AronException("ClassCppWriter", "setupCppClassFromObjectType", "The cpp name of an inner type was empty. Please check the type definition file if all object names are set correctly.");
        }

        std::vector<std::string> namespaces(split);
        namespaces.pop_back();

        CppClassPtr c = std::make_shared<CppClass>(namespaces, split[split.size() - 1]);
        return c;
    }

    void Writer::setupMemberFields(CppClassPtr& c, const serializer::ObjectClassSerializerPtr& o) const
    {
        auto publicFields = o->getPublicVariableDeclarations("");
        for (const auto& f : publicFields)
        {
            c->addPublicField(f);
        }
    }

    void Writer::setupMemberFields(CppClassPtr& c, const serializer::IntEnumClassSerializerPtr& o) const
    {
        auto publicFields = o->getPublicVariableDeclarations("");
        for (const auto& f : publicFields)
        {
            c->addPublicField(f);
        }
    }
}





