/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#pragma once

// STD/STL
#include <memory>
#include <string>

// Base Class
#include <RobotAPI/libraries/aron/core/io/dataIO/reader/ReaderToken.h>

// ArmarX
#include <RobotAPI/libraries/aron/core/Config.h>
#include <RobotAPI/libraries/aron/core/navigator/data/AllNavigators.h>

namespace armarx::aron::dataIO::reader
{
    // BaseClass
    class NavigatorReaderToken;
    typedef std::shared_ptr<NavigatorReaderToken> NavigatorReaderTokenPtr;

    class NavigatorReaderToken :
        virtual public ReaderTokenInterface<datanavigator::NavigatorPtr>
    {
    public:
        using PointerType = NavigatorReaderTokenPtr;

    public:
        // constructors
        NavigatorReaderToken() = delete;
        NavigatorReaderToken(const datanavigator::NavigatorPtr& data)
        {
            descriptor = data->getDescriptor();
            element = data;

            switch (descriptor)
            {
                case data::Descriptor::eDict:
                {
                    allMemberNames = datanavigator::DictNavigator::DynamicCastAndCheck(data)->getAllKeys();
                    childrenSize = data->childrenSize();
                    break;
                }
                case data::Descriptor::eList:
                {
                    childrenSize = data->childrenSize();
                    break;
                }
                default:
                    throw error::DescriptorNotValidException("NavigatorReaderToken", "NavigatorReaderToken", "Received an invalid instance for a reader token.", descriptor);
            }
        }

        data::Descriptor getTypeOfNext() const override
        {
            const datanavigator::NavigatorPtr next = getNextElement();
            return next->getDescriptor();
        }

        datanavigator::NavigatorPtr getNextElement() const override
        {
            if (!hasNextElement())
            {
                return nullptr;
            }

            switch (descriptor)
            {
                case data::Descriptor::eDict:
                {
                    datanavigator::DictNavigatorPtr casted = datanavigator::DictNavigator::DynamicCast(element);
                    datanavigator::NavigatorPtr ret = casted->getElement(getCurrentKey());
                    return ret;
                }
                case data::Descriptor::eList:
                {
                    datanavigator::ListNavigatorPtr casted = datanavigator::ListNavigator::DynamicCast(element);
                    datanavigator::NavigatorPtr ret = casted->getElement(currentIndex);
                    return ret;
                }
                default:
                    throw error::DescriptorNotValidException("NavigatorReaderToken", "getNextElement", "Could not resove a type of a navigator. Allowed are only container types due to performance", descriptor);
            }
        }

        virtual bool hasNextElement() const override
        {
            switch (descriptor)
            {
                case data::Descriptor::eDict:
                {
                    datanavigator::DictNavigatorPtr casted = datanavigator::DictNavigator::DynamicCast(element);
                    datanavigator::NavigatorPtr ret = casted->getElement(getCurrentKey());
                    return (bool) ret;
                }
                case data::Descriptor::eList:
                {
                    datanavigator::ListNavigatorPtr casted = datanavigator::ListNavigator::DynamicCast(element);
                    datanavigator::NavigatorPtr ret = casted->getElement(currentIndex);
                    return (bool) ret;
                }
                default:
                    throw error::DescriptorNotValidException("NavigatorReaderToken", "hasData", "Could not resove a type of a navigator. Allowed are only container types due to performance", descriptor);
            }
        }

    private:

    };
}
